using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

/*
 *  ***********************
 *  *      RarMount2      *
 *  ***********************
 * 
 *  Author JackJson
 * 
 *   Changelog:
 *   2008-02-03  First release
 * 
 *  If you change the source code, please notify me.
 * 
 *  http://forum.team-mediaportal.com/streaming_rar_files-t12448.html
 * 
 */

/*
 * USAGE, mount:
 *  rarmount2 "c:\folder\archive.rar"
 *  imdisk -a -t proxy -m M: -u 5 -o ro,rem,ip -f 127.0.0.1:32123 -b 0 -s 31G
 * Unmount:
 *  imdisk -d -u 5
 * 
 * Additional information:
 * ImDisk:  http://www.ltr-data.se/opencode.html#ImDisk
 * Old RAR: http://www.bsdg.org/SWAG/DATATYPE/0035.PAS.html
 * New RAR: http://darkshire.org/~jhkim/src/rar/technote.txt
 *
 * Abilities:
 *  Mount uncompressed RAR file and stream it's content, for example an AVI
 *
 * Limits:
 *  The RAR archive must be uncompressed (i.e. stored mode)
 *  The RAR archive may only contain one file (fairly easy to fix)
 *  Limited NTFS support, files larger than ~15 GiB are most likely not supported
 *  Requires ImDisk (since I'm not skilled enought to write my own drivers)
 *  Etc, etc, etc
 * 
 * Next release will hopefully:
 *  Find SRT/SUB file in RAR dir and mount it beside the uncompressed RAR content
 */

namespace RarMount2
{
    internal delegate void SetStatusDelegate(String message);

    public partial class frmMain : Form
    {
        private static bool boolDebugMode = true;
        private string subtitleFolder = "";
        public static string[] Args = null;
        private bool boolSilentMode = false;
        private uint intListenPort;
        private string strMainFile = "";
        private string strNtfsHelpFile = "";
        private SimpleNTFS ntfsFileSystem;

        private TcpListener tlListener;
        private Thread listenThread;

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            // Check debug mode setting
            string strDebugMode = getSetting("DEBUG");
            if (!strDebugMode.Equals("True") && !strDebugMode.Equals(""))
                boolDebugMode = false;

            debugPrint("Application started");

            // Read settings from file
            uint.TryParse(getSetting("LISTENPORT"), out intListenPort);
            if (intListenPort == 0)
                intListenPort = 32123; // Default listen port

            // Get NTFS help file
            strNtfsHelpFile = getSetting("NTFSFILE");
            if (strNtfsHelpFile.Length == 0)
                strNtfsHelpFile = "ntfs.bin"; // Default NTFS help file

            // Get subtitle folder
            subtitleFolder = getSetting("SUBTITLEPATH");
            if (subtitleFolder.Length == 0)
                subtitleFolder = @"c:\subtitles"; // Default subtitle folder

            // Check command line arguments
            for (int i = 0; i < Args.Length; i++)
            {
                if (Args[i].ToLower() == "/silent")
                {
                    boolSilentMode = true;
                }
                else
                {
                    // Presume RAR archive
                    strMainFile = Args[i];
                    // FIXME: Break?
                }
            }

            // If not silent mode, show form
            if (!boolSilentMode)
            {
                this.WindowState = FormWindowState.Normal;
                this.ShowInTaskbar = true;
            }
            else
            {
                this.ShowInTaskbar = true;
            }
            
        }

        // Reads setting from XML file
        private string getSetting(string strSettingTitle)
        {
            XmlDocument xmlSettingsDoc = new XmlDocument();
            try
            {
                // Open and read settings file
                xmlSettingsDoc.Load("rarmount_settings.xml");
                return xmlSettingsDoc.GetElementsByTagName(strSettingTitle).Item(0).InnerText.ToString();
            }
            catch (Exception e)
            {
                // Error while reading settings
                debugPrint("DEBUG - Error while reading setting '" + strSettingTitle  + "': " + e.Message.ToString());
                return "";
            }
        }

        // Saves a setting
        private void saveSettings()
        {
            XmlDocument xmlSettingsDoc = new XmlDocument();
            XmlElement xmlSettingsElement = xmlSettingsDoc.CreateElement("SETTINGS");
            XmlElement xmlValueElement;
            string[] strNames = new string[] { "DEBUG", "LISTENPORT", "NTFSFILE", "SUBTITLEPATH" };
            string[] strValues = new string[] { boolDebugMode.ToString(), intListenPort.ToString(), strNtfsHelpFile, subtitleFolder };
            for (int i = 0; i < strNames.Length; i++)
            {
                xmlValueElement = xmlSettingsDoc.CreateElement(strNames[i]);
                xmlValueElement.InnerText = strValues[i];
                xmlSettingsElement.AppendChild(xmlValueElement);
            }
            xmlSettingsDoc.AppendChild(xmlSettingsElement);
            xmlSettingsDoc.Save("rarmount_settings.xml");
            debugPrint("Settings saved");
        }

        internal static void debugPrint(string strDebugMessage)
        {
            // If debug is enabled ...
            if (boolDebugMode)
            {
                // ... write to debug file
                StreamWriter swFile = File.AppendText("rarmount_debug.txt");
                swFile.Write(DateTime.Now.ToString());
                swFile.Write(" - ");
                swFile.WriteLine(strDebugMessage);
                swFile.Close();
            }
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            try
            {
                FileInfo fiMain = new FileInfo(strMainFile);
                if (!fiMain.Exists)
                {
                    setStatus("Archive not found!");
                    debugPrint("Archive not found: " + strMainFile);
                    exitWithDelay();
                }
                else
                {
                    RarCollection rcMain = new RarCollection(strMainFile);
                    if (rcMain.filesLoadedOk())
                    {
                        setStatus(rcMain.getLoadedFilesCount() + " RAR file(s) loaded");
                        debugPrint(rcMain.getLoadedFilesCount() + " RAR file(s) loaded successfully");

                        // Subtitles
                        // Remove old
                        Subtitle.removeOldSubtitles(subtitleFolder);

                        // Find new
                        string subtitle = Subtitle.findSubtitle(strMainFile, rcMain.getStoredFileName());
                        if (subtitle.Length > 0)
                        {
                            string subtitleSource = Path.Combine(fiMain.DirectoryName, subtitle);
                            string subtitleExtension = subtitle.Substring(subtitle.Length - 4, 4);
                            string subtitleTarget = Path.Combine(subtitleFolder, String.Format("UNPACKED{0}", subtitleExtension));
                            debugPrint("Subtitle found: " + subtitle);
                            try
                            {
                                File.Copy(subtitleSource, subtitleTarget, true);
                                debugPrint("Copying subtitle to '" + subtitleTarget + "'.");
                            }
                            catch (Exception ex)
                            {
                                debugPrint("ERROR while copying subtitle! Please make sure the target folder exists: " + subtitleFolder);
                            }
                            
                        }

                        // Init NTFS
                        if (initNTFSfs(rcMain))
                        {
                            runImDiskListner();
                        }
                        else
                        {
                            setStatus("Error while initiating NTFS");
                            debugPrint("Error while initiating simple NTFS file system, exiting");
                        }
                    }
                    else
                    {
                        setStatus("Error while loading RAR file(s)");
                        debugPrint("Error while loading RAR file(s)");
                        exitWithDelay();
                    }
                }
            }
            catch (Exception err)
            {
                setStatus("Error occured!");
                debugPrint("Error occured when mounting " + strMainFile + ": " + err.Message);
                exitWithDelay();
            }
        }

        // Initiate NTFS filesystem
        private bool initNTFSfs(RarCollection rcMain)
        {
            FileInfo fiNtfsFile = new FileInfo(strNtfsHelpFile);

            // Does the help file (default: ntfs.bin) even exist?
            if (!fiNtfsFile.Exists)
            {
                debugPrint("NTFS binary help file not found: " + strNtfsHelpFile);
                return false;
            }
            else if (fiNtfsFile.Length != 2611200)
            {
                debugPrint("NTFS binary help file has wrong file size, expected 2611200, got: " + fiNtfsFile.Length);
                return false;
            }

            ntfsFileSystem = new SimpleNTFS(fiNtfsFile.FullName, rcMain);

            if (ntfsFileSystem.loadedOk())
            {
                debugPrint("NTFS file system initiated successfully");
                return true;
            }
            else
            {
                return false;
            }
        }

        private void ListenForClients()
        {
            this.tlListener.Start();
            setStatus("Waiting for connection ...");
            debugPrint("Listning for connection on port " + intListenPort);

            TcpClient tcClient = tlListener.AcceptTcpClient();
            setStatus("Connected to: " + tcClient.Client.RemoteEndPoint.ToString());
            debugPrint("Accepted request from " + tcClient.Client.RemoteEndPoint.ToString());
            if (tcClient.Connected)
            {
                
                NetworkStream nsStream = tcClient.GetStream();

                //tcClient.ReceiveTimeout = 1000;
                //nsStream.ReadTimeout = 1000;

                byte[] byteSendBuffer;
                byte[] byteFileReadBuffer;

                byte[] byteReadBuffer = new byte[1024];
                int intReceivedDataLength = -1; // <- start value
                while (intReceivedDataLength != 0)
                {
                    try
                    {
                        intReceivedDataLength = nsStream.Read(byteReadBuffer, 0, 1024);

                        if (intReceivedDataLength > 0)
                        {
                            ulong lngResponseCode = BitConverter.ToUInt64(byteReadBuffer, 0);
                            switch (lngResponseCode)
                            {
                                case 0:
                                    //debugPrint("Command: 0 (IMDPROXY_NULL_REQ)");
                                    break;
                                case 1:
                                    setStatus("Connected to ImDisk");
                                    //debugPrint("Command: 1 (IMDPROXY_INFO_REQ)");
                                    ulong disk_size = 33285996544; // 31 GiB
                                    ulong req_alignment = 512;
                                    ulong flags = 0x1; //read only
                                    byteSendBuffer = new byte[sizeof(ulong) * 3];
                                    BitConverter.GetBytes(disk_size).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                    BitConverter.GetBytes(req_alignment).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                    BitConverter.GetBytes(flags).CopyTo(byteSendBuffer, sizeof(ulong) * 2);
                                    //debugPrint("Sending IMDPROXY_INFO_RESP");
                                    nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                    break;
                                case 2:
                                    ulong read_offset = BitConverter.ToUInt64(byteReadBuffer, sizeof(ulong) * 1);
                                    ulong read_length = BitConverter.ToUInt64(byteReadBuffer, sizeof(ulong) * 2);
                                    //debugPrint("Command: 2 (IMDPROXY_READ_REQ), offset: " + read_offset.ToString() + ", length: " + read_length.ToString());
                                    byteFileReadBuffer = new byte[(int)read_length];
                                    int intFileReadBytes = ntfsFileSystem.readData(byteFileReadBuffer, read_offset, (int)read_length);
                                    ulong read_errorno = 0; // =9 if some problem occured
                                    ulong read_resp_length = (ulong)intFileReadBytes;
                                    byteSendBuffer = new byte[sizeof(ulong) * 2];
                                    BitConverter.GetBytes(read_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                    BitConverter.GetBytes(read_resp_length).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                    //debugPrint("Sending IMDPROXY_READ_RESP");
                                    nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length); // ERROR!
                                    nsStream.Write(byteFileReadBuffer, 0, intFileReadBytes);
                                    break;
                                case 3:
                                    //debugPrint("Command: 3 (IMDPROXY_WRITE_REQ)");
                                    // Info: Errorno 9 = EBADF
                                    ulong write_errorno = 9;
                                    ulong write_length = 0;
                                    byteSendBuffer = new byte[sizeof(ulong) * 2];
                                    BitConverter.GetBytes(write_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                    BitConverter.GetBytes(write_length).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                    //debugPrint("Sending IMDPROXY_WRITE_RESP");
                                    nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                    break;
                                case 4:
                                    //debugPrint("Command: 4 (IMDPROXY_CONNECT_REQ)");
                                    ulong connect_errorno = 0;
                                    ulong connect_object_ptr = 0;
                                    byteSendBuffer = new byte[sizeof(ulong) * 2];
                                    BitConverter.GetBytes(connect_errorno).CopyTo(byteSendBuffer, sizeof(ulong) * 0);
                                    BitConverter.GetBytes(connect_object_ptr).CopyTo(byteSendBuffer, sizeof(ulong) * 1);
                                    //debugPrint("Sending IMDPROXY_CONNECT_RESP");
                                    nsStream.Write(byteSendBuffer, 0, byteSendBuffer.Length);
                                    break;
                                default:
                                    debugPrint("Unknow ImDisk command: " + lngResponseCode.ToString());
                                    break;
                            }

                        }

                    }
                    catch (Exception err)
                    {
                        debugPrint("ImDiskListner error, " + err.Message);
                        intReceivedDataLength = 0;
                    }
                }
                nsStream.Close();
            }
            tcClient.Close();

            tlListener.Stop();

            setStatus("Connection lost, exiting ...");
            debugPrint("Connection to ImDisk lost");

            exitWithDelay();
        }

        // Start and run ImDisk listner
        private void runImDiskListner()
        {
            tlListener = new TcpListener(IPAddress.Any, (int)intListenPort);
            listenThread = new Thread(new ThreadStart(ListenForClients));
            listenThread.Start();
        }

        // Set main window status label
        private void setStatus(string strStatusMessage)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new SetStatusDelegate(setStatus), new Object[] { strStatusMessage });
                return;
            }
            lblStatus.Text = strStatusMessage;
            this.Refresh();
        }

        private void exitWithDelay()
        {
            if (this.InvokeRequired)
            {
                this.Invoke(new MethodInvoker(exitWithDelay));
                return;
            }
            // Sleep for 3 seconds (30 * 100 ms)
            // New 2008-05-24: 1 sec is enough
            for (int i = 0; i < 10; i++)
            {
                System.Threading.Thread.Sleep(100);
                Application.DoEvents();
            }
            this.Close();
            //Application.Exit();
        }

        private void frmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            saveSettings();
            debugPrint("Application closed");
        }

    }

    
}