﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace RarMount2
{
    class RarFile
    {
        private string strFileName;  // <- full path to file
        private long lngFileSize;    // <- rar file size
        private FileStream fsReader;

        private long lngRarFileSize;    // <- size of stored file contained in this rar file
        private long lngRarFileStart;   // <- offset to where the real file data begins (i.e. rar header + previous files)
        private ulong lngRarFileOffset; // <- offset in relation to lngRarFileFullSize (ex: if .rar contains the first 100byte of myfile.txt, .r01 will have offset 100)

        // Following variables are only important in first file (.rar, .part001.rar)
        private string strRarFileName;   // <- name of stored file
        private long lngRarFileFullSize; // <- full file size of stored file

        public RarFile(string strLoadFileName, long lngLoadRarFileOffset)
        {
            strFileName = strLoadFileName;
            lngRarFileOffset = (ulong)lngLoadRarFileOffset;
            FileInfo fiFile = new FileInfo(strLoadFileName);
            lngFileSize = fiFile.Length;
        }

        public int readData(long lngPosition, byte[] byteReadBuffer, int intOffset, int intCount)
        {
            fsReader.Position = lngPosition;
            return fsReader.Read(byteReadBuffer, intOffset, intCount);
        }

        public bool loadFileInfo()
        {
            // Open RAR file for reading
            try
            {
                fsReader = new FileStream(strFileName, FileMode.Open, FileAccess.Read);

                // Check RAR file header
                if (!checkFileHeader())
                {
                    // Wrong RAR file header!
                    frmMain.debugPrint("DEBUG - ERROR, archive '" + strFileName + "' has wrong RAR file header!");
                    return false;
                }

                // Read rar archive header size, 2 byte (16 bit) int 
                //  starting at position 12 (7 byte file header + 5 byte unneeded data)
                int intRarHeaderSize = readInt16(12);

                // Read file header flags to determine if compressed file is larger than 2gb
                bool boolLargeFileMode;
                byte[] byteRarFileSize = new byte[8];
                byte[] byteRarFileFullSize = new byte[8];
                int intFileHeaderFlags = readInt8(11 + (long)intRarHeaderSize);
                string strFileHeaderFlags = Convert.ToString(intFileHeaderFlags, 2);
                if (strFileHeaderFlags.Substring(strFileHeaderFlags.Length - 1, 1) == "1")
                    boolLargeFileMode = true; // 64-bit file size
                else
                    boolLargeFileMode = false; // 32-bit file size

                // Read rar archive data starting position (where the stored file begins)
                //lngRarFileStart = (long)readInt16(12 + (long)intRarHeaderSize) + (long)intRarHeaderSize + 7;
                lngRarFileStart = (long)readInt16(-1) + (long)intRarHeaderSize + 7;

                // Read stored file data size (no of bytes stored in current archive, 
                //  i.e. around 15mb for a spanned xvid movie)
                readByte32(-1).CopyTo(byteRarFileSize, 0);

                // Read stored file full size (if extracted, i.e. 700mb for a xvid movie)
                readByte32(-1).CopyTo(byteRarFileFullSize, 0);

                // Read type of compression
                if (readInt8(fsReader.Position + 10) != 0x30) // Skip some unneeded data, then read 1 byte
                {
                    frmMain.debugPrint("DEBUG - ERROR, archive '" + strFileName + "' is compressed!");
                    return false; // <- archive is compressed !
                }

                // Read stored file name size
                int intRarFileNameLength = readInt16(-1);

                fsReader.Position = fsReader.Position + 4; // Skip unneeded data

                // If large file mode (> 2gb), read high 4 bytes of 64-bit file sizes
                if (boolLargeFileMode)
                {
                    readByte32(-1).CopyTo(byteRarFileSize, 4);
                    readByte32(-1).CopyTo(byteRarFileFullSize, 4);
                }
                lngRarFileSize = BitConverter.ToInt64(byteRarFileSize, 0);
                lngRarFileFullSize = BitConverter.ToInt64(byteRarFileFullSize, 0);

                // Read stored file name
                byte[] byteRarReadBuffer = new byte[intRarFileNameLength];
                fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
                strRarFileName = Encoding.ASCII.GetString(byteRarReadBuffer);

                return true; // All ok!
            }
            catch (Exception ex)
            {
                frmMain.debugPrint("DEBUG - Error while loading info from '" + strFileName + "': " + ex.Message);
                return false;
            }
        }

        private int readInt8(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            return fsReader.ReadByte();
        }

        private int readInt16(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[2];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return (int)BitConverter.ToInt16(byteRarReadBuffer, 0);
        }

        private byte[] readByte32(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[4];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return byteRarReadBuffer;
        }

        private long readInt32(long lngPosition)
        {
            if (lngPosition >= 0)
                fsReader.Position = lngPosition;
            byte[] byteRarReadBuffer = new byte[4];
            fsReader.Read(byteRarReadBuffer, 0, byteRarReadBuffer.Length);
            return (long)BitConverter.ToInt32(byteRarReadBuffer, 0);
        }

        public ulong getRarFileOffset()
        {
            return lngRarFileOffset;
        }

        public long getRarFileSize()
        {
            return lngRarFileSize;
        }

        public long getRarFileStart()
        {
            return lngRarFileStart;
        }

        public string getRarFileName()
        {
            return strRarFileName;
        }

        public long getRarFileFullSize()
        {
            return lngRarFileFullSize;
        }

        // Look for valid RAR file header
        private bool checkFileHeader()
        {
            byte[] byteRealRarHeader = new byte[7] { 0x52, 0x61, 0x72, 0x21, 0x1A, 0x07, 0x00 };
            byte[] byteFileHeader = new byte[7];
            fsReader.Position = 0;
            fsReader.Read(byteFileHeader, 0, 7);
            bool boolFileHeaderIsRar = true;
            for (int i = 0; i < 7; i++)
            {
                if (byteFileHeader[i] != byteRealRarHeader[i])
                {
                    boolFileHeaderIsRar = false;
                    break;
                }
            }
            return boolFileHeaderIsRar;
        }
    }
}
